import fastify, { FastifyError, FastifyInstance, FastifyReply, FastifyRequest } from 'fastify'
import helmet from '@fastify/helmet'
import { IncomingMessage, Server, ServerResponse } from 'http'
import { RouterMap } from 'decorators/router'
import { mongoose } from '@typegoose/typegoose'
import logger from 'logger/logger'
import config from 'config/config'
import { ConnectOptions } from 'mongoose'

const zReqParserPlugin = require('plugins/zReqParser')

const zTokenParserPlugin = require('plugins/zTokenParser')

const apiAuthPlugin = require('plugins/apiauth')

const fs = require('fs')
const join = require('path').join

const Redis = require('ioredis')
const RedisStore = require('connect-redis').default

const store = new RedisStore({
  client: new Redis({
    enableAutoPipelining: true,
    host: process.env.REDIS_HOST,
    db: parseInt(process.env.REDIS_DB || '11'),
  }),
  ttl: 600,
})

require('./common/Extend')

export class ApiServer {
  server: FastifyInstance<Server, IncomingMessage, ServerResponse>

  public constructor() {
    this.server = fastify({ logger: true, trustProxy: true })
    this.registerPlugins()
    console.log('version::' + process.version)
  }
  private registerPlugins() {
    this.server.register(require('@fastify/cookie'))
    this.server.register(require('@fastify/session'), {
      cookieName: 'sessionId',
      secret: process.env.COOKIE_SECRET,
      cookie: { secure: false, maxAge: 600000 },
      store,
      expires: 1800,
    })
    this.server.addHook('preHandler', (request, reply, next) => {
      next()
    })
    this.server.register(require('@fastify/formbody'))
    this.server.register(zReqParserPlugin)
    this.server.register(helmet, {
      hidePoweredBy: false,
      contentSecurityPolicy: false,
    })
    this.server.register(zTokenParserPlugin)

    this.server.register(apiAuthPlugin, {
      secret: {
        private: config.api.token_secret_private,
        public: config.api.token_secret_public,
      },
      expiresIn: config.api.token_expiresIn,
    })
    if (process.env.NODE_ENV !== 'production') {
      this.server.register(require('@fastify/cors'), {})
    }

    this.server.register(require('@fastify/view'), {
      engine: {
        ejs: require('ejs'),
      },
    })
  }

  private registerRouter() {
    logger.log('register api routers')
    let self = this
    for (let [controller, config] of RouterMap.decoratedRouters) {
      for (let data of config.data) {
        logger.info(
          'add router',
          data.method || 'all',
          data.path,
          `${data.target.constructor.name}.${controller.name}()`,
        )
        // @ts-ignore
        self.server[data.method || 'all'](
          data.path,
          {
            preValidation: async function (request: FastifyRequest, reply: FastifyReply) {
              request.roles = config.roles
              await this.apiAuth(request, reply)
            },
          },
          controller,
        )
      }
    }
  }
  /**
   * 加载所有的controller
   */
  initControllers() {
    logger.info('Bootstrap controllers...')
    const controllers = join(__dirname, './controllers')
    fs.readdirSync(controllers)
      .filter((file: string) => ~file.search(/^[^.].*\.(ts|js)$/))
      .forEach((file: any) => {
        // logger.log(file);
        return require(join(controllers, file))
      })
  }

  initSchedules() {
    // new DiscordSchedule().scheduleAll()
  }

  initServcers() {
    // new DiscordSvr().init()
  }

  async connectDB() {
    const options: ConnectOptions = {
      minPoolSize: 5,
      maxPoolSize: 10,
      keepAlive: true,
      keepAliveInitialDelay: 300000,
    }
    const uri = config.db_main
    logger.info(`connect to  ${uri} ...`)
    try {
      // await mongoose.createConnection(uri, options)
      await mongoose.connect(uri, options)
      logger.log('DB Connected')
    } catch (err) {
      logger.log(`DB Connection Error: ${err.message}`)
    }
  }
  private setErrHandler() {
    this.server.setNotFoundHandler(function (
      request: any,
      reply: { send: (arg0: { errcode: number; errmsg: string }) => void },
    ) {
      reply.send({ errcode: 404, errmsg: 'page not found' })
    })
    this.server.setErrorHandler(function (error: FastifyError, request: FastifyRequest, reply: FastifyReply) {
      let statusCode = (error && error.statusCode) || 100
      if (statusCode >= 500) {
        logger.error(error)
      } else if (statusCode >= 400) {
        logger.warn(error)
      } else {
        logger.warn(error)
      }
      reply.code(200).send({
        errcode: statusCode,
        errmsg: error ? error.message : 'unknown error',
      })
    })
  }
  /**
   * 格式化接口返回数据, 统一封装成如下格式
   * {
   *     errcode: 0,
   *     errmsg?: '',
   *     data: any
   * }
   * @private
   */
  private setFormatSend() {
    this.server.addHook('preSerialization', async (request: FastifyRequest, reply: FastifyReply, payload) => {
      reply.header('X-Powered-By', 'PHP/5.4.16')
      // @ts-ignore
      if (!payload.errcode) {
        // @ts-ignore
        if (payload.direct) {
          // @ts-ignore
          delete payload.direct
          return payload
        }
        payload = {
          errcode: 0,
          data: payload,
        }
      }
      return payload
    })
  }

  public async start() {
    let self = this
    return new Promise(async (resolve, reject) => {
      await self.connectDB()
      self.initControllers()
      self.registerRouter()
      self.setErrHandler()
      self.setFormatSend()
      self.initSchedules()
      self.initServcers()
      this.server.listen({ port: config.api.port, host: config.api.host }, (err: any, address: any) => {
        if (err) {
          logger.log(err)
          process.exit(0)
        }
        resolve && resolve(address)
      })
    })
  }
}
