import axios from 'axios'
import { hmacsha256, sha1 } from 'utils/security.util'
import crypto from 'crypto'
import { generateKVStr } from 'utils/net.util'
import logger from 'logger/logger'

export function createSimpleSign(data: any) {
  let timestamp = Date.now()
  let appid = process.env.ALCHEMY_APPID
  let secret = process.env.ALCHEMY_APP_SECRET
  let signData = { appid, timestamp }
  signData = Object.assign(signData, data)
  let signStr = Object.keys(signData)
    .sort()
    .map(key => `${key}=${signData[key]}`)
    .join('&')
  let sign = hmacsha256(signStr, secret)
  return {
    appid,
    timestamp,
    sign,
  }
}

export function createSha1Sign() {
  let timestamp = Date.now()
  let appid = process.env.ALCHEMY_APPID
  let secret = process.env.ALCHEMY_APP_SECRET
  let sign = sha1(appid + secret + timestamp)
  return {
    appid,
    timestamp,
    sign,
  }
}

/**
 * Check if the pay result sign is valid
 * @param data - the data to be checked
 * @returns true if the sign is valid, false otherwise
 */
export function checkPayResultSign(data: any) {
  const { appId, orderNo, crypto, network, address, signature } = data
  const sign = hmacsha256(appId + orderNo + crypto + network + address, process.env.ALCHEMY_APP_SECRET)
  return sign === signature
}

export function createPayResultSign(data: any) {
  const { appId, orderNo, crypto, network, address } = data
  const sign = hmacsha256(appId + orderNo + crypto + network + address, process.env.ALCHEMY_APP_SECRET)
  return sign
}

export function checkSimpleSign(headers: any, data: any) {
  // alchemy 很不严谨, 有时候是 appid, 有时候是 appId
  const { appid, appId, timestamp, sign } = headers
  let appIdToCheck = appId || appid
  let signData = { appid: appIdToCheck, timestamp }
  signData = Object.assign(signData, data)
  let signStr = Object.keys(signData)
    .sort()
    .map(key => `${key}=${signData[key]}`)
    .join('&')
  const expectedSign = hmacsha256(signStr, process.env.ALCHEMY_APP_SECRET)
  // const expectedSign = sha1(appIdToCheck + process.env.ALCHEMY_APP_SECRET + timestamp)
  return sign === expectedSign
}

export function checkSha1Sign(headers: any) {
  const { appid, appId, timestamp, sign } = headers
  let appIdToCheck = appId || appid
  const expectedSign = sha1(appIdToCheck + process.env.ALCHEMY_APP_SECRET + timestamp)
  return sign === expectedSign
}

/**
 * Create page sign
 * @param plainText - plain text to be encrypted
 * @returns encrypted text
 */
export function createPageSign(plainText: string) {
  let secret = process.env.ALCHEMY_APP_SECRET
  try {
    const plainTextData = Buffer.from(plainText, 'utf8')
    const secretKey = Buffer.from(secret, 'utf8')
    const iv = secret.substring(0, 16)

    const cipher = crypto.createCipheriv('aes-128-cbc', secretKey, iv)

    let encrypted = cipher.update(plainTextData)
    encrypted = Buffer.concat([encrypted, cipher.final()])

    return encrypted.toString('base64')
  } catch (e) {
    console.log(`AES encrypting exception, msg is ${e.toString()}`)
  }
  return null
}

export function checkPageSign(params: any) {
  const keySet = new Set(['appId', 'address', 'callbackUrl'])
  let signData = {}
  for (let k in params) {
    if (keySet.has(k)) {
      signData[k] = params[k]
    }
  }

  let signStr = generateKVStr({ data: signData, sort: true })
  let expectedSign = createPageSign(signStr)
  logger.info(`signStr is ${signStr}, expectedSign is ${expectedSign}, origin sign is ${params.sign}`)
  let { sign } = params
  return sign === expectedSign
}
