import { ZError } from 'common/ZError'
import { singleton } from 'decorators/singleton'

import { Client, Events, GatewayIntentBits, Guild } from 'discord.js'
import { AuthRecordClass } from 'modules/AuthRecord'
import { DocumentType } from '@typegoose/typegoose'

const DISCORD_API_HOST = 'https://discord.com/api/v10'
const DISCORD_APP_HOST = 'https://discordapp.com/api'

export async function exchangeDiscrodCodeForToken(code: string) {
  const clientId = process.env.DISCORD_CLIENT_ID
  const clientSecret = process.env.DISCORD_CLIENT_SECRET
  const redirectUri = 'https://oauth-svr.cebggame.com/discord/redirect_uri'

  const response = await fetch('https://discord.com/api/oauth2/token', {
    method: 'POST',
    headers: {
      'Content-Type': 'application/x-www-form-urlencoded',
    },
    body: new URLSearchParams({
      grant_type: 'authorization_code',
      client_id: clientId,
      client_secret: clientSecret,
      redirect_uri: redirectUri,
      code,
    }),
  })

  const data = await response.json()
  return data
}

export async function getAvableAccessToken(record: DocumentType<AuthRecordClass>) {
  if (record.expiresIn < Date.now() / 1000 + 60 * 5) {
    console.log('discord access token expired')
    const data = await refreshDiscordToken(record.refreshToken)
    if (data.code) {
      throw new ZError(30, 'refresh token error')
    }
    record.accessToken = data.access_token
    record.refreshToken = data.refresh_token
    record.expiresIn = (Date.now() / 1000 + data.expires_in || 0) | 0
    await record.save()
  }
  return record.accessToken
}
export async function refreshDiscordToken(rtoken: string) {
  const clientId = process.env.DISCORD_CLIENT_ID
  const clientSecret = process.env.DISCORD_CLIENT_SECRET
  const response = await fetch('https://discord.com/api/oauth2/token', {
    method: 'POST',
    headers: {
      'Content-Type': 'application/x-www-form-urlencoded',
    },
    body: new URLSearchParams({
      grant_type: 'refresh_token',
      client_id: clientId,
      client_secret: clientSecret,
      refresh_token: rtoken,
    }),
  })
  const data = await response.json()
  return data
}
/**
 * 用户基本信息
 * @param token
 * @returns
 */
export async function userInfo(token: string) {
  const response = await fetch(`${DISCORD_APP_HOST}/users/@me`, {
    headers: {
      authorization: `Bearer ${token}`,
    },
  })

  const data = await response.json()
  return data
}
/**
 * 用户所在的工会列表, oauth2方式
 * @param token
 * @returns
 */
export async function userGuildList(token: string) {
  const url = `${DISCORD_APP_HOST}/users/@me/guilds`
  const response = await fetch(url, {
    headers: {
      authorization: `Bearer ${token}`,
    },
  })
  const data = await response.json()
  return data
}

/**
 * 用户在某个工会的信息, oauth2方式
 * @param token
 * @param guildId
 * @returns
 * 返回的数据结构:
 * https://discord.com/developers/docs/resources/guild#guild-member-object
 */
export async function userGuildMember(token: string, guildId: string) {
  const url = `${DISCORD_APP_HOST}/users/@me/guilds/${guildId}/member`
  const response = await fetch(url, {
    headers: {
      authorization: `Bearer ${token}`,
    },
  })
  const data = await response.json()
  return data
}

/**
 * 直接通过discord接口, 获取工会中某用户信息
 * 返回的数据结构:
 * https://discord.com/developers/docs/resources/guild#guild-member-object
 */
export async function getGuildMember(uid: string) {
  const url = `${DISCORD_API_HOST}/guilds/${process.env.DISCROD_GUILD_ID}/members/${uid}`
  const token = process.env.DISCORD_BOT_TOKEN
  const response = await fetch(url, {
    headers: {
      authorization: `Bot ${token}`,
    },
  })
  const data = await response.json()
  return data
}

@singleton
export class DiscordSvr {
  private client: Client
  private guild: Guild
  public async init() {
    console.log('DiscordSvr init')
    this.client = new Client({ intents: [GatewayIntentBits.Guilds, GatewayIntentBits.GuildMembers] })
    this.client.once(Events.ClientReady, async c => {
      console.log(`Ready! Logged in as ${c.user.tag}`)
      this.guild = await this.client.guilds.fetch(process.env.DISCROD_GUILD_ID)
      console.log('guild', this.guild)
      await this.guild.members.fetch()
      for (const member of this.guild.members.cache.values()) {
        console.log(member.user)
      }
    })
    this.client.on(Events.GuildMemberUpdate, async (oldMember, newMember) => {
      console.log('GuildMemberUpdate', oldMember, newMember)
      // Check if the member's roles have changed
      // if (oldMember.roles.cache.size !== newMember.roles.cache.size) {
      //   // Get the added and removed roles
      //   const addedRoles = newMember.roles.cache.filter(role => !oldMember.roles.cache.has(role.id))
      //   const removedRoles = oldMember.roles.cache.filter(role => !newMember.roles.cache.has(role.id))
      //   console.log('addedRoles', addedRoles)
      //   console.log('removedRoles', removedRoles)
      // }
    })
    this.client.on(Events.GuildMemberAdd, async member => {
      console.log('GuildMemberAdd', member)
    })
    this.client.on(Events.GuildMemberRemove, async member => {
      console.log('GuildMemberRemove', member)
    })
    this.client.login(process.env.DISCORD_BOT_TOKEN)
  }

  public async updateCache() {
    console.log('update discord cache')
    if (!this.guild) {
      throw new ZError(10, 'DiscordSvr not init')
    }
    await this.guild.members.fetch()
    console.log('update discord cache done')
  }

  public checkUser(uid: string) {
    if (!this.guild) {
      throw new ZError(10, 'DiscordSvr not init')
    }
    for (const member of this.guild.members.cache.values()) {
      console.log(member.user)
    }
    return this.guild.members.cache.has(uid)
  }

  public checkUserRole(uid: string) {
    if (!this.guild) {
      throw new ZError(10, 'DiscordSvr not init')
    }
    // const member = await this.guild.members.fetch({ user: uid, force: true })
    const member = this.guild.members.cache.get(uid)
    if (!member) {
      return false
    }
    return member.roles.cache.has(process.env.DISCORD_ROLE_ID)
  }
}
